/*
 * Copyright (c) 1997 Hong Hunsoo of (KAIST)
 * The Korea Advanced Institute of Science and Technology
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Neither the name of the KAIST nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * The end of copyright
 * The start of EXPLANATIONS:
 *
 * Hong Hunsoo of (KAIST)
 * The Korea Advanced Institute of Science and Technology changed original
 * source code to make it more compact.  I think my modification will be
 * useful to you.
 *
 * By my modification, in X-window, to draw a card in Window, you only need
 * to call card_init(Display *,int screeenno); then
 * extern Pixmap cards[suit][rank] will contain the pattern of the card of
 * suit suit and rank rank. If you do not need that pixmaps any more, you
 * can call card_free();
 *
 * Remeber that cards[4][0] contains the backface of the cards.  I intended
 * to make cards[4][1] contains the pattern of Joker card, but I had no time
 * to do, that is cards[4][1] contains nothing currently.
 *
 * I think cards.c and cards.h can be linked to any other program easily.
 * Therefore I think my modification was a sort of modularization.
 *
 * 1997/8/26
 * Hong Hunsoo
 * department of physics
 * Korea Advanced Institute of Science and Technology
 */
/*
 * Spider
 *
 * (c) Copyright 1989, Donald R. Woods and Sun Microsystems, Inc. (c) Copyright
 * 1990, David Lemke and Network Computing Devices Inc.
 *
 * See copyright.h for the terms of the copyright.
 *
 * @(#)gfx.c    2.1     90/04/25
 *
 */

#include <stdio.h>
#include <stdlib.h>

#include "common.h"
#include "cards.h"

/* card info*/

#define	FACECARD_WIDTH	47
#define	FACECARD_HEIGHT	92

#define	RANK_WIDTH	9
#define	RANK_HEIGHT	14

#define	RANK_LOC_X	4
#define	RANK_LOC_Y	7

#define	SMALL_LOC_X	4
#define	SMALL_LOC_Y	(RANK_HEIGHT + RANK_LOC_Y + 3)

#define	MID_CARD_X	(CARD_WIDTH/2)
#define	MID_CARD_Y	(CARD_HEIGHT/2)

#define	CARD_COL1_X	(3 * CARD_WIDTH/10)
#define	CARD_COL2_X	(CARD_WIDTH/2)
#define	CARD_COL3_X	(7 * CARD_WIDTH/10)

/*
 * 5 diff rows for the two main columns
 * 1 and 5 are top and bottom, 3 is the middle
 * 2 & 4 are for the 10 & 9
 */

#define	CARD_ROW1_Y	(CARD_HEIGHT/5)
#define	CARD_ROW2_Y	(2 * CARD_HEIGHT/5)
#define	CARD_ROW3_Y	(CARD_HEIGHT/2)
#define	CARD_ROW4_Y	(CARD_HEIGHT - 2 * CARD_HEIGHT/5)
#define	CARD_ROW5_Y	(CARD_HEIGHT - CARD_HEIGHT/5)

/* between 1 & 3, 3 & 5 */
#define	CARD_SEVEN_Y	(7 * CARD_HEIGHT/20)
#define	CARD_EIGHT_Y	(CARD_HEIGHT - 7 * CARD_HEIGHT/20)

/* between rows 1 & 2, 4 & 5 */
#define	CARD_TEN_Y1	(3 * CARD_HEIGHT/10)
#define	CARD_TEN_Y2	(CARD_HEIGHT - 3 * CARD_HEIGHT/10)

#include <X11/Xmu/Drawing.h>

#define	ROUND_W	7
#define	ROUND_H	7

static GC cardgc;		/* gc in use when drawing cards */

/* substitue gray1 for Green on mono */

#define gray1_width 16
#define gray1_height 16
static char gray1_bits[] = {
	0x55, 0x55, 0xaa, 0xaa, 0x55, 0x55, 0xaa, 0xaa, 0x55, 0x55, 0xaa, 0xaa,
	0x55, 0x55, 0xaa, 0xaa, 0x55, 0x55, 0xaa, 0xaa, 0x55, 0x55, 0xaa, 0xaa,
	0x55, 0x55, 0xaa, 0xaa, 0x55, 0x55, 0xaa, 0xaa
};

/* logo for card backs */
/* any size logo can be handled.  */

#define logo_width 64
#define logo_height 64

static char logo_bits[]={
0x77,0xd5,0xd7,0xdd,0x77,0xd5,0xd7,0xdd,0xbb,0xea,0xae,0xbb,0xbb,0xea,0xae,
0xbb,0x5d,0x75,0x5d,0x77,0x5d,0x75,0x5d,0x77,0xee,0xba,0xba,0xee,0xae,0xba,
0xba,0xee,0x77,0x5d,0x75,0xdd,0x77,0x5d,0x75,0xdd,0xbb,0xae,0xea,0xba,0xbb,
0xae,0xea,0xba,0x5d,0x57,0xd5,0x75,0x5d,0x57,0xd5,0x75,0xae,0xab,0xaa,0xeb,
0xae,0xab,0xaa,0xeb,0xd7,0x55,0x55,0xd7,0xd7,0x51,0x55,0xd7,0xeb,0xae,0xaa,
0xae,0xeb,0xa2,0xaa,0xae,0x75,0x5d,0x55,0x5d,0x75,0x44,0x15,0x5d,0xba,0xbb,
0xea,0xba,0xba,0x88,0x8a,0xba,0x5d,0x77,0x75,0x75,0x1d,0x11,0x45,0x75,0xee,
0xee,0xba,0xea,0x2e,0x22,0xa2,0xea,0xd7,0xdd,0x5d,0xd5,0x57,0x44,0x10,0xd5,
0xab,0xbb,0xef,0xaa,0xab,0x88,0x88,0xaa,0x57,0x77,0x77,0xd5,0x57,0x11,0x45,
0xd5,0xae,0xee,0xba,0xea,0xae,0x22,0xa2,0xea,0x5d,0xdd,0x5d,0x75,0x5d,0x45,
0x10,0x75,0xba,0xba,0xef,0xba,0xba,0x8a,0x88,0xba,0x75,0x75,0x77,0x5d,0x75,
0x15,0x45,0x5d,0xea,0xea,0xba,0xae,0xea,0x2a,0xa2,0xae,0xdd,0xd5,0x5d,0x77,
0xdd,0x55,0x50,0x77,0xae,0xab,0xaf,0xeb,0xae,0xab,0xa8,0xeb,0xd7,0x55,0x57,
0xd7,0xd7,0x55,0x55,0xd7,0xeb,0xaa,0xaa,0xae,0xeb,0xaa,0xaa,0xae,0x75,0x57,
0x55,0x5d,0x75,0x57,0x55,0x5d,0xba,0xae,0xea,0xba,0xba,0xae,0xaa,0xba,0xdd,
0x5d,0x75,0x75,0xdd,0x5d,0x75,0x75,0xae,0xbb,0xba,0xea,0xae,0xbb,0xba,0xea,
0x77,0x77,0x5d,0xd5,0x77,0x77,0x5d,0xd5,0xeb,0xee,0xee,0xaa,0xeb,0xee,0xae,
0xaa,0xd7,0xdd,0x77,0xd5,0xd7,0xdd,0x77,0xd5,0xae,0xbb,0xbb,0xea,0xae,0xbb,
0xbb,0xea,0x5d,0x77,0x5d,0x75,0x5d,0x77,0x5d,0x75,0xba,0xee,0xee,0xba,0xba,
0xee,0xae,0xba,0x75,0xdd,0x77,0x5d,0x75,0xdd,0x77,0x5d,0xea,0xba,0xbb,0xae,
0xea,0xba,0xbb,0xae,0xd5,0x75,0x5d,0x57,0xd5,0x75,0x5d,0x57,0xaa,0xeb,0xae,
0xab,0xaa,0xeb,0xae,0xab,0x55,0xd7,0xd7,0x55,0x55,0xd7,0xd7,0x55,0xaa,0xae,
0xeb,0xae,0xaa,0xae,0xeb,0xae,0x55,0x5d,0x75,0x5d,0x55,0x5d,0x75,0x5d,0xea,
0xba,0xba,0xbb,0xea,0xba,0xba,0xbb,0x75,0x75,0x5d,0x77,0x75,0x75,0x5d,0x77,
0xba,0xea,0xee,0xee,0xba,0xea,0xee,0xee,0x5d,0xd5,0xd7,0xdd,0x5d,0xd5,0xd7,
0xdd,0xef,0xaa,0xab,0xbb,0xef,0xaa,0xab,0xbb,0x77,0xd5,0x57,0x77,0x77,0xd5,
0x57,0x77,0xba,0xea,0xae,0xee,0xba,0xea,0xae,0xee,0x5d,0x75,0x5d,0xdd,0x5d,
0x75,0x5d,0xdd,0xef,0xba,0xba,0xba,0xef,0xba,0xba,0xba,0x77,0x5d,0x75,0x75,
0x77,0x5d,0x75,0x75,0xba,0xae,0xea,0xea,0xba,0xae,0xea,0xea,0x5d,0x77,0xdd,
0xd5,0x5d,0x77,0xdd,0xd5,0xaf,0xeb,0xae,0xab,0xaf,0xeb,0xae,0xab,0x57,0xd7,
0xd7,0x55,0x57,0xd7,0xd7,0x55,0xaa,0xae,0xeb,0xaa,0xaa,0xae,0xeb,0xaa,0x55,
0x5d,0x75,0x57,0x55,0x5d,0x75,0x57,0xaa,0xba,0xba,0xae,0xaa,0xba,0xba,0xae,
0x75,0x75,0xdd,0x5d,0x55,0x75,0xdd,0x5d,0xba,0xea,0xae,0xbb,0xba,0xea,0xae,
0xbb,0x5d,0xd5,0x77,0x77,0x5d,0xd5,0x77,0x77,0xae,0xaa,0xeb,0xee,0xae,0xaa,
0xeb,0xee};

#include "bitmaps/rank.bm"
#include "bitmaps/face.bm"
#include "bitmaps/suit.bm"

/*
 * Following pixmaps are temporary pixmaps, they will be freed.
 */
Pixmap rank_map[NUM_RANKS], rank_r_map[NUM_RANKS];
Pixmap rank_map_red[NUM_RANKS], rank_r_map_red[NUM_RANKS];

Pixmap suit_map[NUM_SUITS], suit_r_map[NUM_SUITS];
Pixmap suit_sm_map[NUM_SUITS], suit_sm_r_map[NUM_SUITS];

Pixmap suit_lg_map[NUM_SUITS];
Pixmap jack_map[NUM_SUITS];
Pixmap queen_map[NUM_SUITS];
Pixmap king_map[NUM_SUITS];

static GC cardgc_red;
static GC cardgc_black, cardgc_blackline;
static GC cardgc_white;
static GC cardgc_back;

static int back_delta_x, back_delta_y;	/* how much to modify the TS origin */
static unsigned long cardpixel_black, cardpixel_white, cardpixel_green;
static int has_colors;
static int cardscreen;
static Window cardrootwin;
static Display *carddpy;

Pixmap	map_green;
Pixmap	map_red;
Pixmap	**cards;	/* Only this variable can be accessible */
Pixmap	pixmap_joker;	/* == cards[4][1] */
Pixmap	pixmap_back;	/* == cards[4][0] */

static Pixmap make_red_map(char *bits, int width, int height);
static void draw_pip(Pixmap target, Suit suit, int x, int y);
static void draw_did(Pixmap target, Suit suit, int x, int y);
static void draw_center_pip(Pixmap target, Suit suit, int x, int y);
static void draw_two_pips(Pixmap target, Suit suit, int x, int y);
static void draw_four_pips(Pixmap target, Suit suit, int x, int y);
static void draw_six_pips(Pixmap target, Suit suit, int x, int y);
static void draw_eight_pips(Pixmap target, Suit suit, int x, int y);
static void draw_jack(Pixmap target, Suit suit, int x, int y);
static void draw_queen(Pixmap target, Suit suit, int x, int y);
static void draw_king(Pixmap target, Suit suit, int x, int y);
static void draw_rank(Pixmap target, int x, int y, Rank rank, Suit suit);

static void flip_bits(unsigned char *src, unsigned char *dst, int W, int H);
static void rot_180(unsigned char *src, unsigned char *dst, int W, int H);
static void make_backface(Pixmap target);
static void free_nousemaps(void);
static void make_card_maps(void);
static void make_real_cards(void);
static void paint_card(Pixmap one, int x, int y, Rank rank, Suit suit);

void
card_init(Display *d, int scr)
{
	XGCValues gcv;
	long    gcflags;
	XColor  color;
	Colormap cmap;
	unsigned long redpixel;
	int	card_inited = 0;
	/* save these off */

	if (card_inited) {
		abort();
	} else {
		card_inited++;
	}

	carddpy = d;
	cardscreen = scr;
	cardrootwin = RootWindow(carddpy, cardscreen);

	if (DisplayCells(carddpy, cardscreen) > 2) {
		has_colors = True;
	} else {
		has_colors = False;
	}

	cardpixel_black = BlackPixel(carddpy, cardscreen);
	cardpixel_white = WhitePixel(carddpy, cardscreen);

	gcv.foreground = cardpixel_white;
	gcv.background = cardpixel_black;
	gcv.graphics_exposures = False;

	gcflags = GCForeground | GCBackground | GCGraphicsExposures;
	cardgc_white = XCreateGC(carddpy, cardrootwin,
		gcflags, &gcv);

	gcv.foreground = cardpixel_black;
	gcv.background = cardpixel_white;
	gcv.fill_style = FillTiled;

	gcflags = GCForeground | GCBackground |
		GCGraphicsExposures | GCFillStyle;

	cardgc_black = XCreateGC(carddpy, cardrootwin, gcflags, &gcv);

	gcv.foreground = cardpixel_black;
	gcv.background = cardpixel_white;
	gcv.line_width = 0;

	cardgc_blackline = XCreateGC(carddpy, cardrootwin,
		GCForeground | GCBackground | GCLineWidth, &gcv);

	if (has_colors) {
		cmap = DefaultColormap(carddpy, cardscreen);

		color.flags = DoRed | DoGreen | DoBlue;

		/*
		 * color levels are the NeWS RGB values
		 */
		color.red = 13107;	/* 0.2 */
		color.green = 52428;	/* 0.8 */
		color.blue = 39321;	/* 0.6 */

		XAllocColor(carddpy, cmap, &color);
		cardpixel_green = color.pixel;

		color.red = 52428;	/* 0.8 */
		color.green = color.blue = 0;
		XAllocColor(carddpy, cmap, &color);
		redpixel = color.pixel;

		gcv.foreground = redpixel;
		gcv.background = cardpixel_white;
		gcv.fill_style = FillTiled;
		gcflags = GCForeground | GCBackground
			| GCFillStyle | GCGraphicsExposures;

		cardgc_red = XCreateGC(carddpy, cardrootwin,
			gcflags, &gcv);

	} else {
		map_green = XCreateBitmapFromData(carddpy,
			cardrootwin,
			gray1_bits, gray1_width, gray1_height);

		map_red = XCreateBitmapFromData(carddpy,
			cardrootwin,
			gray1_bits, gray1_width, gray1_height);

		gcv.tile = map_red;
		gcv.fill_style = FillTiled; /* XXX */

		gcv.foreground = cardpixel_black;
		gcv.background = cardpixel_white;

		gcflags = GCTile | GCForeground | GCBackground |
			GCFillStyle | GCGraphicsExposures;

		cardgc_red = XCreateGC(carddpy, cardrootwin, gcflags, &gcv);
	}
	make_card_maps();
	make_real_cards();
	return;
}

void
card_free(void)
{
	int	i, j;

	for (i= 0; i< NUM_SUITS; i++) {
		for (j=0; j < NUM_RANKS; j++)
			XFreePixmap(carddpy, cards[i][j]);
		free(cards[i]);
	}
	free(cards[i]); /* cards[4] free */
	free(cards);
	XFreePixmap(carddpy, cards[4][0]); /* backface of the cards */

	XFreeGC(carddpy, cardgc_red);
	XFreeGC(carddpy, cardgc_black);
	XFreeGC(carddpy, cardgc_blackline);
	XFreeGC(carddpy, cardgc_white);
	XFreeGC(carddpy, cardgc_back);

	return;
}
	

static void
make_real_cards(void)
{
	int	i, j;

	cards = (Pixmap **) malloc(sizeof(Pixmap *) * (NUM_SUITS+2));
	if (cards == NULL)
		err("%s(): ", __F);

	for (i=0; i < NUM_SUITS; i++) {
		cards[i] = (Pixmap *) malloc(sizeof(Pixmap) * (NUM_RANKS+1));
		if (cards[i] == NULL)
			err("%s(): ", __F);
		for (j=0; j< NUM_RANKS; j++) {
			cards[i][j] = XCreatePixmap(carddpy,
				cardrootwin, CARD_WIDTH, CARD_HEIGHT,
				DefaultDepth(carddpy, cardscreen));
			paint_card(cards[i][j], 0, 0, j, i);
		}
		cards[i][j] = 0;
	}

	cards[i] = (Pixmap *) malloc(sizeof(Pixmap) * 3);
	pixmap_back = cards[i][0] = XCreatePixmap(carddpy,
		cardrootwin, CARD_WIDTH, CARD_HEIGHT,
		DefaultDepth(carddpy, cardscreen));
	make_backface(cards[i][0]);
	free_nousemaps();
	return;
}

static void
make_backface(Pixmap target)
{
	Colormap cmap;
	GC      logogc;
	Pixmap  tmpmap, map_logo;
	XGCValues gcv;
	XColor	color;
	long    gcflags;

	tmpmap = XCreateBitmapFromData(carddpy, cardrootwin,
		logo_bits, logo_width, logo_height);

	map_logo = XCreatePixmap(carddpy, cardrootwin,
		logo_width, logo_height, DefaultDepth(carddpy, cardscreen));

	back_delta_x = (CARD_WIDTH - logo_width) / 2;
	back_delta_y = (CARD_HEIGHT - logo_height) / 2;

	gcflags = 0;
	if (has_colors) {
		cmap = DefaultColormap(carddpy, cardscreen);
		XAllocNamedColor(carddpy, cmap, "Sea Green", &color, &color);

		gcv.foreground = color.pixel;
		gcv.background = cardpixel_white;
		gcflags = GCForeground | GCBackground;

		logogc = XCreateGC(carddpy, cardrootwin, gcflags, &gcv);
		XCopyPlane(carddpy, tmpmap, map_logo, logogc, 0, 0,
			   logo_width, logo_height, 0, 0, 1);
		XFreeGC(carddpy, logogc);
	} else {
		XCopyPlane(carddpy, tmpmap, map_logo, cardgc_white, 0, 0,
			   logo_width, logo_height, 0, 0, 1);
	}

	XFreePixmap(carddpy, tmpmap);

	gcv.tile       = map_logo; /* XXX */
	gcv.fill_style = FillTiled;
	gcv.graphics_exposures = False;
	gcflags = GCForeground | GCBackground |
		GCTile | GCFillStyle | GCGraphicsExposures;
	cardgc_back = XCreateGC(carddpy, cardrootwin, gcflags, &gcv);
	XSetTSOrigin(carddpy, cardgc_back, back_delta_x, back_delta_y);

	XFillRectangle(carddpy, target, cardgc_back,
		0, 0, CARD_WIDTH, CARD_HEIGHT);
	return;
}

static void
free_nousemaps(void)
{
	int	i;

	for (i=0; i < 4; i++) {
		XFreePixmap(carddpy, rank_map[i]);
		XFreePixmap(carddpy, rank_r_map[i]);
		XFreePixmap(carddpy, rank_map_red[i]);
		XFreePixmap(carddpy, rank_r_map_red[i]);
		XFreePixmap(carddpy, suit_map[i]);
		XFreePixmap(carddpy, suit_r_map[i]);
		XFreePixmap(carddpy, suit_sm_map[i]);
		XFreePixmap(carddpy, suit_sm_r_map[i]);
		if (i == 0)
			XFreePixmap(carddpy, suit_lg_map[i]);
		XFreePixmap(carddpy, jack_map[i]);
		XFreePixmap(carddpy, queen_map[i]);
		XFreePixmap(carddpy, king_map[i]);
	}
	return;
}

static void
make_card_maps(void)
{
	unsigned char *new_bits;
	Rank    r;
	int     i;
	char	*num_bits[4], *jack_bits[4], *queen_bits[4], *king_bits[4];
	int	num_width[4], jack_width[4], queen_width[4], king_width[4];
	int	num_height[4], jack_height[4], queen_height[4], king_height[4];
	char	*num_sm_bits[4];
	int	num_sm_width[4], num_sm_height[4];

	for (r = RANK_ACE; r <= RANK_KING; r++) {
		rank_map[(int) r] = XCreateBitmapFromData(carddpy,
			cardrootwin,
			rank_bits[(int) r], rank_width, rank_height);
		if (rank_map[r] == 0)
			err("%s():" __F);

		new_bits = (unsigned char *)
			calloc(sizeof(rank_bits[(int) r]), 1);
		rot_180((unsigned char *) rank_bits[(int) r], new_bits,
			rank_width, rank_height);
		rank_r_map[(int) r] = XCreateBitmapFromData(carddpy,
			cardrootwin,
			new_bits, rank_width, rank_height);
		if (rank_r_map[r] == 0)
			err("%s():" __F);
		free((char *) new_bits);
	}

	for (r = RANK_ACE; r <= RANK_KING; r++) {
		new_bits = (unsigned char *)
			calloc(sizeof(rank_bits[(int) r]), 1);
		rot_180((unsigned char *) rank_bits[(int) r], new_bits,
			rank_width, rank_height);

		if (has_colors) {
			rank_map_red[(int) r] = XCreateBitmapFromData(carddpy,
				cardrootwin,
				rank_bits[(int) r], rank_width, rank_height);
			if (rank_map_red[r] == 0)
				err("%s():" __F);

			rank_r_map_red[(int) r] = XCreateBitmapFromData(carddpy,
				cardrootwin,
				new_bits, rank_width, rank_height);
			if (rank_r_map_red[r] == 0)
				err("%s():" __F);
		} else {
			rank_map_red[(int) r] = make_red_map((char *)
				rank_bits[(int) r],
				rank_width, rank_height);

			rank_r_map_red[(int) r] =
				make_red_map((char *) new_bits,
				rank_width, rank_height);
		}
		free((char *) new_bits);
	}

	num_bits[0] = spade_bits;
	num_bits[1] = heart_bits;
	num_bits[2] = diamond_bits;
	num_bits[3] = club_bits;

	num_width[0] = spade_width;   num_height[0] = spade_height;
	num_width[1] = heart_width;   num_height[1] = heart_height;
	num_width[2] = diamond_width; num_height[2] = diamond_height;
	num_width[3] = club_width;    num_height[3] = club_height;

	num_sm_bits[0] = spade_sm_bits;
	num_sm_bits[1] = heart_sm_bits;
	num_sm_bits[2] = diamond_sm_bits;
	num_sm_bits[3] = club_sm_bits;

	num_sm_width[0] = spade_sm_width;  num_sm_height[0] = spade_sm_height;
	num_sm_width[1] = heart_sm_width;  num_sm_height[1] = heart_sm_height;
	num_sm_width[2] = diamond_sm_width;num_sm_height[2] = diamond_sm_height;
	num_sm_width[3] = club_sm_width;   num_sm_height[3] = club_sm_height;

	jack_bits[0] = jack_s_bits;
	jack_bits[1] = jack_h_bits;
	jack_bits[2] = jack_d_bits;
	jack_bits[3] = jack_c_bits;

	jack_width[0] = jack_s_width;   jack_height[0] = jack_s_height;
	jack_width[1] = jack_h_width;   jack_height[1] = jack_h_height;
	jack_width[2] = jack_d_width;   jack_height[2] = jack_d_height;
	jack_width[3] = jack_c_width;   jack_height[3] = jack_c_height;

	queen_bits[0] = queen_s_bits;
	queen_bits[1] = queen_h_bits;
	queen_bits[2] = queen_d_bits;
	queen_bits[3] = queen_c_bits;

	queen_width[0] = queen_s_width;   queen_height[0] = queen_s_height;
	queen_width[1] = queen_h_width;   queen_height[1] = queen_h_height;
	queen_width[2] = queen_d_width;   queen_height[2] = queen_d_height;
	queen_width[3] = queen_c_width;   queen_height[3] = queen_c_height;

	king_bits[0] = king_s_bits;
	king_bits[1] = king_h_bits;
	king_bits[2] = king_d_bits;
	king_bits[3] = king_c_bits;

	king_width[0] = king_s_width;   king_height[0] = king_s_height;
	king_width[1] = king_h_width;   king_height[1] = king_h_height;
	king_width[2] = king_d_width;   king_height[2] = king_d_height;
	king_width[3] = king_c_width;   king_height[3] = king_c_height;

	for (i=0; i < 4; i++) {
		/* make all the card bitmaps */
		suit_map[i] = XCreateBitmapFromData(carddpy,
			cardrootwin,
			num_bits[i], num_width[i], num_height[i]);

		new_bits = (unsigned char *) malloc(strlen(num_bits[i])+1);
		flip_bits((unsigned char *) num_bits[i], new_bits, num_width[i],
			 num_height[i]);
		suit_r_map[i] = XCreateBitmapFromData(carddpy,
			cardrootwin,
			new_bits, num_width[i], num_height[i]);
		free((char *) new_bits);


		suit_sm_map[i] = XCreateBitmapFromData(carddpy,
			cardrootwin,
			num_sm_bits[i], num_sm_width[i], num_sm_height[i]);
		new_bits = (unsigned char *) malloc(strlen(num_sm_bits[i]) + 1);
		flip_bits((unsigned char *) num_sm_bits[i], new_bits,
			num_sm_width[i], num_sm_height[i]);

		suit_sm_r_map[i] = XCreateBitmapFromData(carddpy,
			cardrootwin,
			new_bits, num_sm_width[i], num_sm_height[i]);
		free((char *) new_bits);

		if (i == 0) {
			suit_lg_map[i] = XCreateBitmapFromData(carddpy,
				cardrootwin,
				spade_lg_bits, spade_lg_width, spade_lg_height);
		} else {
			suit_lg_map[i] = suit_map[i];
		}

		jack_map[i] = XCreateBitmapFromData(carddpy,
			cardrootwin,
			jack_bits[i], jack_width[i], jack_height[i]);

		queen_map[i] = XCreateBitmapFromData(carddpy,
			cardrootwin,
			queen_bits[i], queen_width[i], queen_height[i]);

		king_map[i] = XCreateBitmapFromData(carddpy,
			cardrootwin,
			king_bits[i], king_width[i], king_height[i]);
	}
	return;
}

/*
 * make a 'red' pixmap by setting the clipmask to the desired shape and
 * pushing 'red' through it
 */
static  Pixmap
make_red_map(char *bits, int width, int height)
{
	Pixmap  tmpmap, newmap;
	static GC cleargc = (GC) 0;
	XGCValues xgcv;

	if (cleargc == (GC) 0) {
		xgcv.function = GXclear;
		cleargc = XCreateGC(carddpy, cardrootwin,
			GCFunction, &xgcv);
	}
	tmpmap = XCreateBitmapFromData(carddpy, cardrootwin,
		bits, width, height);

	newmap = XCreatePixmap(carddpy, cardrootwin,
		width, height, 1);

	/* clear pixmap */
	XFillRectangle(carddpy, newmap, cleargc, 0, 0, width, height);

	XSetClipMask(carddpy, cardgc_red, tmpmap);
	XFillRectangle(carddpy, newmap, cardgc_red, 0, 0, width, height);
	XSetClipMask(carddpy, cardgc_red, None);
	XFreePixmap(carddpy, tmpmap);

	return (newmap);
}
/*
 * delta is 0 if the card is fully showing
 */
static void
paint_card(Pixmap target, int x, int y, Rank rank, Suit suit)
{
	if (suit == SUIT_SPADE || suit == SUIT_CLUB) {
		cardgc = cardgc_black;
	} else {
		cardgc = cardgc_red;
	}

	XFillRectangle(carddpy, target, cardgc_white, x, y,
		CARD_WIDTH, CARD_HEIGHT);

	switch (rank) {
		case RANK_KING:
			draw_king(target, suit, x, y);
			break;
		case RANK_QUEEN:
			draw_queen(target, suit, x, y);
			break;
		case RANK_JACK:
			draw_jack(target, suit, x, y);
			break;

		case RANK_TEN:
			draw_pip(target, suit, MID_CARD_X + x, CARD_TEN_Y1 + y);
			draw_did(target, suit, MID_CARD_X + x, CARD_TEN_Y2 + y);
			draw_eight_pips(target, suit, x, y);
			break;

		case RANK_NINE:
			draw_pip(target, suit, x + MID_CARD_X, y + MID_CARD_Y);
			draw_eight_pips(target, suit, x, y);
			break;

		case RANK_EIGHT:
			draw_did(target, suit, x+MID_CARD_X, y + CARD_EIGHT_Y);
			/* fall thru */
		case RANK_SEVEN:
			draw_pip(target, suit, x+MID_CARD_X, CARD_SEVEN_Y + y);
			/* fall thru */
		case RANK_SIX:
			draw_six_pips(target, suit, x, y);
			break;

		case RANK_FIVE:
			draw_pip(target, suit, x + MID_CARD_X, y + MID_CARD_Y);
			/* fall thru */
		case RANK_FOUR:
			draw_four_pips(target, suit, x, y);
			break;

		case RANK_THREE:
			draw_pip(target, suit, x + MID_CARD_X, y + MID_CARD_Y);
			/* fall thru */
		case RANK_DEUCE:
			draw_two_pips(target, suit, x, y);
			break;
		case RANK_ACE:
			draw_center_pip(target, suit,
				x + MID_CARD_X, y + MID_CARD_Y);
			break;
		default:
			err("%s(): wrong %dth line of %s file",
				__F, __L, __FILE__);
	}

	draw_rank(target, x, y, rank, suit);
	XDrawRectangle(carddpy, target, cardgc_black, x, y,
		CARD_WIDTH-1, CARD_HEIGHT-1);
	return;
}

/*
 * NOTE -- for all the pip drawers except the one that actually plots the
 * bits, the location is the card's location.  the drawer's take the pip's
 * center as location.
 */

/*
 * draws right-side-up pip
 *
 * location is for center of pip
 */
static void
draw_pip(Pixmap target, Suit suit, int x, int y)
{
	int     w, h;

	switch (suit) {
		case SUIT_SPADE:
			w = spade_width;
			h = spade_height;
			break;
		case SUIT_DIAMOND:
			x++;
			w = diamond_width;
			h = diamond_height;
			break;
		case SUIT_HEART:
			y++;
			w = heart_width;
			h = heart_height;
			break;
		case SUIT_CLUB:
			y++;
			w = club_width;
			h = club_height;
			break;
		default:
			err("%s(): wrong %dth line of %s file",
				__F, __L, __FILE__);
	}
	XCopyPlane(carddpy, suit_map[suit], target, cardgc,
		   0, 0, w, h, x - w / 2, y - h / 2, 1);
}

/*
 * draws upside-down pip
 *
 * location is for center of pip
 */
static void
draw_did(Pixmap target, Suit suit, int x, int y)
{
	int     w, h;

	switch (suit) {
	case SUIT_SPADE:
		w = spade_width;
		h = spade_height;
		break;
	case SUIT_DIAMOND:
		x++;
		w = diamond_width;
		h = diamond_height;
		break;
	case SUIT_HEART:
		y++;
		w = heart_width;
		h = heart_height;
		break;
	case SUIT_CLUB:
		y++;
		w = club_width;
		h = club_height;
		break;
	default:
		err("%s(): wrong %dth line of %s file", __F, __L, __FILE__);
	}
	XCopyPlane(carddpy, suit_r_map[suit], target, cardgc,
		   0, 0, w, h, x - w / 2, y - h / 2, 1);
}

/*
 * draws big center pip
 */
static void
draw_center_pip(Pixmap target, Suit suit, int x, int y)
{
	int     w, h;

	switch (suit) {
	case SUIT_SPADE:
		w = spade_lg_width;
		h = spade_lg_height;
		break;
	case SUIT_DIAMOND:
		w = diamond_width;
		h = diamond_height;
		break;
	case SUIT_HEART:
		w = heart_width;
		h = heart_height;
		break;
	case SUIT_CLUB:
		w = club_width;
		h = club_height;
		break;
	default:
		err("%s(): wrong %dth line of %s file", __F, __L, __FILE__);
	}
	XCopyPlane(carddpy, suit_lg_map[suit], target, cardgc,
		   0, 0, w, h,
		   x - w / 2, y - h / 2, 1);
}

/*
 * draw_two_pips
 */
static void
draw_two_pips(Pixmap target, Suit suit, int x, int y)
{
	draw_pip(target, suit, x + MID_CARD_X, y + CARD_ROW1_Y);
	draw_did(target, suit, x + MID_CARD_X, y + CARD_ROW5_Y);
}

/*
 * draw_four_pips
 */
static void
draw_four_pips(Pixmap target, Suit suit, int x, int y)
{
	draw_pip(target, suit, x + CARD_COL1_X, y + CARD_ROW1_Y);
	draw_did(target, suit, x + CARD_COL1_X, y + CARD_ROW5_Y);

	draw_pip(target, suit, x + CARD_COL3_X, y + CARD_ROW1_Y);
	draw_did(target, suit, x + CARD_COL3_X, y + CARD_ROW5_Y);
}

static void
draw_six_pips(Pixmap target, Suit suit, int x, int y)
{
	draw_pip(target, suit, x + CARD_COL1_X, y + CARD_ROW1_Y);

	draw_pip(target, suit, x + CARD_COL3_X, y + CARD_ROW1_Y);


	/* these are only visible when its not clipped */
	draw_pip(target, suit, x + CARD_COL1_X, y + CARD_ROW3_Y);
	draw_did(target, suit, x + CARD_COL1_X, y + CARD_ROW5_Y);

	draw_pip(target, suit, x + CARD_COL3_X, y + CARD_ROW3_Y);
	draw_did(target, suit, x + CARD_COL3_X, y + CARD_ROW5_Y);
}

static void
draw_eight_pips(Pixmap target, Suit suit, int x, int y)
{
	draw_pip(target, suit, x + CARD_COL1_X, y + CARD_ROW1_Y);

	draw_pip(target, suit, x + CARD_COL3_X, y + CARD_ROW1_Y);

	/* these are only visible when its not clipped */
	draw_pip(target, suit, x + CARD_COL1_X, y + CARD_ROW2_Y);
	draw_did(target, suit, x + CARD_COL1_X, y + CARD_ROW4_Y);
	draw_did(target, suit, x + CARD_COL1_X, y + CARD_ROW5_Y);

	draw_pip(target, suit, x + CARD_COL3_X, y + CARD_ROW2_Y);
	draw_did(target, suit, x + CARD_COL3_X, y + CARD_ROW4_Y);
	draw_did(target, suit, x + CARD_COL3_X, y + CARD_ROW5_Y);
}


static void
draw_jack(Pixmap target, Suit suit, int x, int y)
{
	XCopyPlane(carddpy, jack_map[suit], target, cardgc,
		   0, 0, FACECARD_WIDTH, FACECARD_HEIGHT,
		   x + (CARD_WIDTH - FACECARD_WIDTH) / 2,
		   y + (CARD_HEIGHT - FACECARD_HEIGHT) / 2, 1);

	XDrawRectangle(carddpy, target, cardgc,
		       x + (CARD_WIDTH - FACECARD_WIDTH) / 2,
		       y + (CARD_HEIGHT - FACECARD_HEIGHT) / 2,
		       FACECARD_WIDTH, FACECARD_HEIGHT);
}

static void
draw_queen(Pixmap target, Suit suit, int x, int y)
{
	XCopyPlane(carddpy, queen_map[suit], target, cardgc,
		   0, 0, FACECARD_WIDTH, FACECARD_HEIGHT,
		   x + (CARD_WIDTH - FACECARD_WIDTH) / 2,
		   y + (CARD_HEIGHT - FACECARD_HEIGHT) / 2, 1);

	XDrawRectangle(carddpy, target, cardgc,
		       x + (CARD_WIDTH - FACECARD_WIDTH) / 2,
		       y + (CARD_HEIGHT - FACECARD_HEIGHT) / 2,
		       FACECARD_WIDTH, FACECARD_HEIGHT);
}

static void
draw_king(Pixmap target, Suit suit, int x, int y)
{
	XCopyPlane(carddpy, king_map[suit], target, cardgc,
		   0, 0, FACECARD_WIDTH, FACECARD_HEIGHT,
		   x + (CARD_WIDTH - FACECARD_WIDTH) / 2,
		   y + (CARD_HEIGHT - FACECARD_HEIGHT) / 2, 1);

	XDrawRectangle(carddpy, target, cardgc,
		       x + (CARD_WIDTH - FACECARD_WIDTH) / 2,
		       y + (CARD_HEIGHT - FACECARD_HEIGHT) / 2,
		       FACECARD_WIDTH, FACECARD_HEIGHT);
}

static void
draw_rank(Pixmap target, int x, int y, Rank rank, Suit suit)
{
	int     w, h;
	if (suit == SUIT_HEART || suit == SUIT_DIAMOND) {
		XCopyPlane(carddpy, rank_map_red[rank], target, cardgc,
			0, 0, RANK_WIDTH, RANK_HEIGHT,
			x + RANK_LOC_X, y + RANK_LOC_Y, 1);

		/* reversed rank? */
		XCopyPlane(carddpy, rank_r_map_red[rank], target, cardgc,
			0, 0, RANK_WIDTH, RANK_HEIGHT,
			x + (CARD_WIDTH - RANK_WIDTH - RANK_LOC_X),
			y + (CARD_HEIGHT - RANK_HEIGHT - RANK_LOC_Y), 1);
	} else {
		XCopyPlane(carddpy, rank_map[rank], target, cardgc,
			0, 0, RANK_WIDTH, RANK_HEIGHT,
			x + RANK_LOC_X, y + RANK_LOC_Y, 1);

		XCopyPlane(carddpy, rank_r_map[rank], target, cardgc,
			0, 0, RANK_WIDTH, RANK_HEIGHT,
			x + (CARD_WIDTH - RANK_WIDTH - RANK_LOC_X),
			y + (CARD_HEIGHT - RANK_HEIGHT - RANK_LOC_Y), 1);
	}

	switch (suit) {
	case SUIT_SPADE:
		w = spade_sm_width;
		h = spade_sm_height;
		break;
	case SUIT_HEART:
		w = heart_sm_width;
		h = heart_sm_height;
		break;
	case SUIT_DIAMOND:
		x++;		/* offset the smaller width */
		w = diamond_sm_width;
		h = diamond_sm_height;
		break;
	case SUIT_CLUB:
		w = club_sm_width;
		h = club_sm_height;
		break;
	default:
		err("%s(): wrong %dth line of %s file", __F, __L, __FILE__);
	}
	XCopyPlane(carddpy, suit_sm_map[suit], target, cardgc,
		   0, 0, w, h,
		   x + SMALL_LOC_X, y + SMALL_LOC_Y, 1);

	/* reversed suit ? */
	XCopyPlane(carddpy, suit_sm_r_map[suit], target, cardgc,
		0, 0, w, h,
		x + (CARD_WIDTH - w - SMALL_LOC_X),
		y + (CARD_HEIGHT - h - SMALL_LOC_Y), 1);
}

static unsigned char _reverse_byte[0x100] = {
	0x00, 0x80, 0x40, 0xc0, 0x20, 0xa0, 0x60, 0xe0,
	0x10, 0x90, 0x50, 0xd0, 0x30, 0xb0, 0x70, 0xf0,
	0x08, 0x88, 0x48, 0xc8, 0x28, 0xa8, 0x68, 0xe8,
	0x18, 0x98, 0x58, 0xd8, 0x38, 0xb8, 0x78, 0xf8,
	0x04, 0x84, 0x44, 0xc4, 0x24, 0xa4, 0x64, 0xe4,
	0x14, 0x94, 0x54, 0xd4, 0x34, 0xb4, 0x74, 0xf4,
	0x0c, 0x8c, 0x4c, 0xcc, 0x2c, 0xac, 0x6c, 0xec,
	0x1c, 0x9c, 0x5c, 0xdc, 0x3c, 0xbc, 0x7c, 0xfc,
	0x02, 0x82, 0x42, 0xc2, 0x22, 0xa2, 0x62, 0xe2,
	0x12, 0x92, 0x52, 0xd2, 0x32, 0xb2, 0x72, 0xf2,
	0x0a, 0x8a, 0x4a, 0xca, 0x2a, 0xaa, 0x6a, 0xea,
	0x1a, 0x9a, 0x5a, 0xda, 0x3a, 0xba, 0x7a, 0xfa,
	0x06, 0x86, 0x46, 0xc6, 0x26, 0xa6, 0x66, 0xe6,
	0x16, 0x96, 0x56, 0xd6, 0x36, 0xb6, 0x76, 0xf6,
	0x0e, 0x8e, 0x4e, 0xce, 0x2e, 0xae, 0x6e, 0xee,
	0x1e, 0x9e, 0x5e, 0xde, 0x3e, 0xbe, 0x7e, 0xfe,
	0x01, 0x81, 0x41, 0xc1, 0x21, 0xa1, 0x61, 0xe1,
	0x11, 0x91, 0x51, 0xd1, 0x31, 0xb1, 0x71, 0xf1,
	0x09, 0x89, 0x49, 0xc9, 0x29, 0xa9, 0x69, 0xe9,
	0x19, 0x99, 0x59, 0xd9, 0x39, 0xb9, 0x79, 0xf9,
	0x05, 0x85, 0x45, 0xc5, 0x25, 0xa5, 0x65, 0xe5,
	0x15, 0x95, 0x55, 0xd5, 0x35, 0xb5, 0x75, 0xf5,
	0x0d, 0x8d, 0x4d, 0xcd, 0x2d, 0xad, 0x6d, 0xed,
	0x1d, 0x9d, 0x5d, 0xdd, 0x3d, 0xbd, 0x7d, 0xfd,
	0x03, 0x83, 0x43, 0xc3, 0x23, 0xa3, 0x63, 0xe3,
	0x13, 0x93, 0x53, 0xd3, 0x33, 0xb3, 0x73, 0xf3,
	0x0b, 0x8b, 0x4b, 0xcb, 0x2b, 0xab, 0x6b, 0xeb,
	0x1b, 0x9b, 0x5b, 0xdb, 0x3b, 0xbb, 0x7b, 0xfb,
	0x07, 0x87, 0x47, 0xc7, 0x27, 0xa7, 0x67, 0xe7,
	0x17, 0x97, 0x57, 0xd7, 0x37, 0xb7, 0x77, 0xf7,
	0x0f, 0x8f, 0x4f, 0xcf, 0x2f, 0xaf, 0x6f, 0xef,
	0x1f, 0x9f, 0x5f, 0xdf, 0x3f, 0xbf, 0x7f, 0xff
};
#define S(x,y) src[(H-1-(y))*W+(x)]
#define D(x,y) dst[(H-1-(y))*W+(x)]

static void
flip_bits(unsigned char *src, unsigned char *dst, int W, int H)
{
	int     x, y;
	W = (W + 7) / 8;
	for (y = 0; y < H; y++) {
		for (x = 0; x < W; x++) {
			D(x, y) = S(x, H - 1 - y);
		}
	}
}

static void
rot_180(unsigned char *src, unsigned char *dst, int W, int H)
{
	int     x, y;
	int     width = W;
	unsigned char *new;
	int     bit;
	W = (W + 7) / 8;
	for (y = 0; y < H; y++) {
		for (x = 0; x < W; x++) {
			D(x, y) = _reverse_byte[S(W - 1 - x, H - 1 - y)];
		}
	}

	/* shift it over */
	new = (unsigned char *) calloc((unsigned) W * H, (unsigned) 1);
	for (y = 0; y < H; y++) {
		for (x = 0; x < W * 8; x++) {
			bit = (*(dst + (x + (W * 8 - width)) / 8 + y * W)
			       & (1 << ((x + (W * 8 - width)) % 8))) ? 1 : 0;
			*(new + x / 8 + y * W) = (bit << (x % 8)) |
				(*(new + x / 8 + y * W) & ~(1 << (x % 8)));
		}
	}
	memcpy((char *) dst, (char *) new, W * H);
	free((char *) new);
}

void
set_gclinewidth100th(GC gc, const int linewidth)
{
	static int saved_linewidth = -100;
	static GC saved_gc;
	XGCValues xgcv;

	if ((saved_gc == gc && saved_linewidth == linewidth) || linewidth < 0)
		return;
	saved_linewidth = xgcv.line_width = linewidth;
	XChangeGC(carddpy, gc, GCLineWidth, &xgcv);
	saved_gc = gc;
	return;
}

void
paint_cardback(Pixmap target, int x, int y)
{
	/*
	 * change the origin so cards will have the same back anywhere on the
	 * table there should be a tile centered in the card, with the
	 * surrounding tiles being partial
	 */
	XSetTSOrigin(carddpy, cardgc_back, x + back_delta_x, y + back_delta_y);

	XmuFillRoundedRectangle(carddpy, target, cardgc_back,
		x + 1, y + 1, CARD_WIDTH - 2, CARD_HEIGHT - 2,
		ROUND_W, ROUND_H);

	set_gclinewidth100th(cardgc_blackline, 2);
	XmuDrawRoundedRectangle(carddpy, target, cardgc_blackline,
		x + 1, y + 1, CARD_WIDTH - 2, CARD_HEIGHT - 2,
		ROUND_W, ROUND_H);
	return;
}

void
put_card(Pixmap target, int suit, int rank, int x, int y)
{
	GC	thisgc;

	switch (suit) {
		case SUIT_SPADE: case SUIT_CLUB:
			thisgc = cardgc_black;
			break;
		case SUIT_HEART: case SUIT_DIAMOND:
			thisgc = cardgc_red;
			break;
		case 4:
			thisgc = cardgc_back;
			break;
	}
	if (suit < 4) {
		set_gclinewidth100th(cardgc_blackline, 1);
		XSetTile(carddpy, thisgc, cards[suit][rank]);
	} else {
		set_gclinewidth100th(cardgc_blackline, 2);
	}
	XSetTSOrigin(carddpy, thisgc, x, y);

	XmuFillRoundedRectangle(carddpy, target, thisgc,
		x, y, CARD_WIDTH-1, CARD_HEIGHT-1, ROUND_W, ROUND_H);

	XmuDrawRoundedRectangle(carddpy, target, cardgc_blackline,
		x, y, CARD_WIDTH-1, CARD_HEIGHT-1, ROUND_W, ROUND_H);

	return;
}
