/*
 * Copyright (c) 1997 Hong Hunsoo of (KAIST)
 * The Korea Advanced Institute of Science and Technology
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Neither the name of the KAIST nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>

#include "common.h"
#include "f_fonts.h"

int	counts[STROKE_NO];
struct point_t points[1024];

struct point_t *savedp;
int	firstcalled;
int	dotindex;
int	stroke_cnt;

static void font_sorting(struct font_t *font);

struct char_t *
character(void)
{
	static int size;
	static int firstguest = 1;
	struct char_t *ret;

	if (firstguest) {
		firstguest = 0;
		size = sizeof(struct char_t);
	}
	ret = (struct char_t *) malloc(size);
	if (ret == NULL)
		err("%s(): while malloc of char_t", __F);

	memset(ret, 0x00, size);
	return ret;
}

struct char_t *
char_init(char *name, int enc, int dw, int three, int four)
{
	struct char_t *cp;

	cp = (struct char_t *) malloc(sizeof(struct char_t));
	if (cp == NULL)
		err("%s(): while malloc of char_t of %s", __F, name);

	cp->c_charname = name;
	cp->c_encoding = enc;
	cp->c_dwidth = dw;
	cp->c_bbox.b_height = three;
	cp->c_bbox.b_offy = four;

	return cp;
}


void
closepath(void)
{
	points[dotindex].x = savedp->x;
	points[dotindex].y = savedp->y;

	stroke_cnt++;
	dotindex++;

	counts[stroke_cnt] = dotindex;
}

void
ah(int x, int y)
{

	if (firstcalled == 1) { /* first stroke of a character */
		savedp = &points[0];
		firstcalled = 0;
		dotindex = 0;
		stroke_cnt = 0;
		counts[0] = 0;
	}
	savedp = &points[dotindex];

	points[dotindex].x = x; points[dotindex].y = -y;
	dotindex++;
}
void
lt(int x, int y)
{
	points[dotindex].x = x;
	points[dotindex].y = -y;
	dotindex++;
}


void
lth(int x, int y)
{
	points[dotindex].x = x;
	points[dotindex].y = -y;
	dotindex++;
}

int	thru_cth;

void
ct(int x1, int y1, int x2, int y2, int x3, int y3)
{
	static int ct_inited = 1;
	static int segno, segno3, *s0p, *s1p, *s2p, *s3p;
	int	i, x0, y0, xt, yt, delta;

	if (ct_inited) {
		ct_inited = 0;
		segno = 10;
		segno3 = segno*segno*segno;

		s0p = (int *) malloc(sizeof(int) * (segno+1));
		s1p = (int *) malloc(sizeof(int) * (segno+1));
		s2p = (int *) malloc(sizeof(int) * (segno+1));
		s3p = (int *) malloc(sizeof(int) * (segno+1));

		if (!(s0p && s1p && s2p && s3p))
			err("%s(): one of sp's == NULL", __F);
		for (i=0; i <= segno; i++) {
			delta = (segno) - i;
			s0p[i] = delta*delta*delta;
			s1p[i] = 3*delta*delta*i;
			s2p[i] = 3*delta*i*i;
			s3p[i] = i*i*i;
		}
	}

	x0 = points[dotindex-1].x;
	y0 = points[dotindex-1].y;
	y1 *= -1;
	y2 *= -1;
	y3 *= -1;


	for (i=1; i<= (segno); i++, dotindex++) {
		xt = x0 * s0p[i] + x1 * s1p[i] + x2 * s2p[i] + x3 * s3p[i];
		yt = y0 * s0p[i] + y1 * s1p[i] + y2 * s2p[i] + y3 * s3p[i];

		xt /= segno3;
		yt /= segno3;
		points[dotindex].x = xt;
		points[dotindex].y = yt;
	}
	thru_cth = 0;
}

void
cth(int x1, int y1, int x2, int y2, int x3, int y3)
{
	thru_cth = 1;
	ct(x1, y1, x2, y2, x3, y3);
}

void
endchar(struct char_t *cp)
{
	struct point_t *ptp;
	int	i, *ip, size;

	size = dotindex * sizeof(struct point_t);
	ptp = (struct point_t *) malloc(size);
	if (ptp == NULL)
		err("%s(): ptp == NULL", __F);
	memcpy(ptp, points, size);
	cp->c_coors = ptp;
	cp->c_ncoors = dotindex;

	ip = cp->c_strokes;

	for (i=0; i <= stroke_cnt; i++)
		ip[i] = counts[i];

	if (stroke_cnt != cp->c_nstrokes) {
		fprintf(stderr, "cp->c_charname=%s\n", cp->c_charname);
		err("%s(): stroke_cnt(%d) != (cp->c_nstrokes(%d))",
			__F, stroke_cnt, cp->c_nstrokes);
	}

	/* Hmm */
	dotindex = 0;
	stroke_cnt = 0;
	counts[0] = 0;
	stroke_cnt = 0;
	return;
}

void
put_char(struct char_t *cp)
{
	int	i, j, *stp;
	struct point_t *pp;

	stp = cp->c_strokes;
	pp = cp->c_coors;
	printf("%s\n", cp->c_charname);

	for (i=0; i < cp->c_nstrokes; i++) {
		for (j= stp[i]; j < stp[i+1]; j++) {
			printf("%6d %6d\n", pp[j].x, pp[j].y);
		}
	}
	return;
}

struct font_t *
assemble(func_t *funclist)
{
	func_t *p;
	struct font_t *fp;
	struct char_t **cpp0, **cpp1;
	int	cnt, i;

	for (cnt = 0, p = funclist; p[0]; p++)
		cnt++;

	fp = (struct font_t *) malloc(sizeof(struct font_t));
	if (fp == NULL)
		err("%s(): fp == NULL", __F);
	fp->f_nchars = cnt;

	cpp0 = (struct char_t **) malloc(sizeof(struct char_t *) * cnt);
	if (cpp0 == NULL)
		err("%s(): cpp0 == NULL [0]", __F);
	fp->f_chars = cpp0;

	cpp1 = (struct char_t **) malloc(sizeof(struct char_t *) * cnt);
	if (cpp1 == NULL)
		err("%s(): cpp1 == NULL [1]", __F);
	fp->f_chars0 = cpp1;

	for (i=0, p = &funclist[0]; p[0]; p++, i++) {
		cpp1[i] = cpp0[i] = (p[0])();
	}
	font_sorting(fp);
	return fp;
}

int
method_byname(struct char_t **this, struct char_t **that)
{
	return strcmp((*this)->c_charname, (*that)->c_charname);
}

int
method_byencoding(struct char_t **this, struct char_t **that)
{
	int	one, two;
	one = (*this)->c_encoding, two = (*that)->c_encoding;

	return (one > two ? 1 : one == two ? 0 : -1);
}

static void
font_sorting(struct font_t *font)
{
	int	(*method)();

	method = method_byname;
	qsort(font->f_chars0, font->f_nchars, sizeof(struct char_t *), method);

	method = method_byencoding;
	qsort(font->f_chars, font->f_nchars, sizeof(struct char_t *), method);
}

struct char_t *
bsearch4charbyname(struct font_t *font, const char *target)
{
	struct char_t **base, **p;
	int lim, cmp;
	char	*ret;

	base = font->f_chars0; /* XXX */

	for (lim =  font->f_nchars; lim != 0; lim >>= 1) {
		p = base + (lim >> 1);
		ret = p[0]->c_charname;
		cmp = strcmp(target, ret);

		if (cmp == 0)
			return p[0];
		if (cmp > 0) {  /* key > p: move right */
			base = (p + 1);
			lim--;
		} /* else move left */
	}
	fprintf(stderr, "%s(): can not find a character whose name is %s\n",
		__F, target);
	return NULL;
}

struct char_t *
bsearch4charbyencoding(struct font_t *font, int target)
{
	struct char_t **base, **p;
	int lim, cmp;
	int	ret;

	base = font->f_chars; /* XXX */

	for (lim =  font->f_nchars; lim != 0; lim >>= 1) {
		p = base + (lim >> 1);
		ret = p[0]->c_encoding;
		cmp = ret < target ? 1 : ret > target ? -1 : 0;

		if (cmp == 0)
			return p[0];
		if (cmp > 0) {  /* key > p: move right */
			base = (p + 1);
			lim--;
		} /* else move left */
	}
	fprintf(stderr, "%s(): can not find a character whose encoding is %d\n",
		__F, target);
	return NULL;
}
