/*
 * Copyright (c) 1997 Hong Hunsoo of (KAIST)
 * The Korea Advanced Institute of Science and Technology
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Neither the name of the KAIST nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <X11/Xlib.h>

#include "common.h"
#include "xcommon.h"
#include "f_fonts.h"
#include "f_inter.h"

XPoint xpoints[1024];
struct char_t *chars[80];

void
xput_char(Display *dpy, Window win, GC line_gc, int x, int y,
	int scale, char ch, struct font_t *fp)
{
	int	i, last, *stp, *dots;
	struct char_t *cp;
	struct point_t *src;
	XPoint *xp;

	cp = bsearch4charbyencoding(fp, ch);
	if (cp == NULL)
		return;
	src = cp->c_coors;
	xp = xpoints;

	for (i=0, last = cp->c_ncoors; i < last; i++, xp++, src++) {
		xp->x = (src->x * scale) / UNIT + x;
		xp->y = (src->y * scale) / UNIT + y;
	}

	/* xp and src has changed by incrementation operator */

	stp = cp->c_strokes;
	dots = cp->c_strokes;
	for (i=0, xp = xpoints, last = cp->c_nstrokes;i < last; i++) {
		XDrawLines(dpy, win, line_gc, &xp[stp[i]], stp[i+1]-stp[i],
			CoordModeOrigin);
	}

	return;
}

void
xput_char_t(Display *dpy, Window win, GC line_gc, int x, int y,
	int scale, struct char_t *cp)
{
	int	i, last, *stp, *dots;
	struct point_t *src;
	XPoint *xp;

	src = cp->c_coors;
	xp = xpoints;

	for (i=0, last = cp->c_ncoors; i < last; i++, xp++, src++) {
		xp->x = (src->x * scale) / UNIT + x;
		xp->y = (src->y * scale) / UNIT + y;
	}

	/* xp and src has changed by incrementation operator */

	stp = cp->c_strokes;
	dots = cp->c_strokes;
	for (i=0, xp = xpoints, last = cp->c_nstrokes;i < last; i++) {
		XDrawLines(dpy, win, line_gc, &xp[stp[i]], stp[i+1]-stp[i],
			CoordModeOrigin);
	}

	return;
}

void
xfill_char_t(Display *dpy, Window win, GC fill_gc, int x, int y,
	int scale, struct char_t *cp)
{
	int	i, last, *stp, *dots;
	struct point_t *src;
	XPoint *xp;

	src = cp->c_coors;
	xp = xpoints;

	for (i=0, last = cp->c_ncoors; i < last; i++, xp++, src++) {
		xp->x = (src->x * scale) / UNIT + x;
		xp->y = (src->y * scale) / UNIT + y;
	}

	/* xp and src has changed by incrementation operator */

	stp = cp->c_strokes;
	dots = cp->c_strokes;
	for (i=0, xp = xpoints, last = cp->c_nstrokes;i < last; i++) {
		XFillPolygon(dpy, win, fill_gc, &xp[stp[i]], stp[i+1]-stp[i],
			Complex, CoordModeOrigin);
	}

	return;
}

void
xstr_put(Display *dpy, Window win, GC line_gc, int x, int y,
	int scale, char *chp, struct font_t *fp)
{
	struct char_t *cp;

	for (; chp[0]; chp++) {
		cp = bsearch4charbyencoding(fp, chp[0]);
		if (cp == NULL)
			continue;

		xput_char_t(dpy, win, line_gc, x, y, scale, cp);
		x += (cp->c_dwidth * scale) / UNIT;
	}
	return;
}

void
xstr_cput(Display *dpy, Window win, GC line_gc, int width, int y,
	int scale, char *chp, struct font_t *fp)
{
	struct char_t *cp, **dst;
	int	x, span;

	for(dst = chars, span = 0; chp[0]; chp++, dst++) {
		cp = bsearch4charbyencoding(fp, chp[0]);
		if (cp == NULL)
			continue;
		*dst = cp;
		span += (cp->c_dwidth * scale)/UNIT;

	}
	*dst = NULL;

	x = (width-span)/2;
	for (dst = chars; dst[0]; dst++) {
		xput_char_t(dpy, win, line_gc, x, y, scale, dst[0]);
		x += (dst[0]->c_dwidth * scale) / UNIT;
	}
	return;
}

void
xstr_cputandfill(Display *dpy, Window win, GC line_gc, GC fill_gc,
	int width, int y, int scale, char *chp, struct font_t *fp)
{
	struct char_t *cp, **dst;
	int	x, span;

	for(dst = chars, span = 0; chp[0]; chp++, dst++) {
		cp = bsearch4charbyencoding(fp, chp[0]);
		if (cp == NULL)
			continue;
		*dst = cp;
		span += (cp->c_dwidth * scale)/UNIT;

	}
	*dst = NULL;

	x = (width-span)/2;

	XSetFunction(dpy, fill_gc, GXxor);
	for (dst = chars; dst[0]; dst++) {
		xfill_char_t(dpy, win, fill_gc, x, y, scale, dst[0]);
		x += (dst[0]->c_dwidth * scale) / UNIT;
	}
	XSetFunction(dpy, fill_gc, GXcopy);

	x = (width-span)/2;
	for (dst = chars; dst[0]; dst++) {
		xput_char_t(dpy, win, line_gc, x, y, scale, dst[0]);
		x += (dst[0]->c_dwidth * scale) / UNIT;
	}

	return;
}
