#include <stdio.h>
#include "X11/Xlib.h"
#include "common.h"
#include "xcommon.h"

unsigned long
name_to_pixel(Colormap colorm, const char *name)
{
	XColor	exact, returned;

	XAllocNamedColor(dpy, colorm, name,
		&returned, &exact);
	return returned.pixel;
}


int
error(Display *disp, XErrorEvent *event)
{
	char	errtext[64];

	printf("%4d %4d %4d |%ld|----", event->error_code,
		event->request_code, event->minor_code, event->resourceid);
	XGetErrorText(disp, event->error_code, errtext, 64);
	printf("%s\n", errtext);
	abort();
	return 1;
}

void
set_gclinewidth(GC gc, const int linewidth)
{
	unsigned long mask = GCLineWidth;
	XGCValues       xgcv;

	if (linewidth < 0)
		return;
	xgcv.line_width = linewidth;
	XChangeGC(dpy, gc, mask, &xgcv);
	return;
}


char *
get_eventname(const int eventno)
{
	static char *names[]={
		"ZERO", "FIRST",
		"KeyPress", "KeyRelease",
		"ButtonPress", "ButtonRelease",
		"MotionNotify", "EnterNotify",
		"LeaveNotify", "FocusIn",
		"FocusOut", "KeymapNotify",
		"Expose", "GraphicsExpose",
		"NoExpose", "VisibilityNotify",
		"CreateNotify", "DestroyNotify",
		"UnmapNotify", "MapNotify",
		"MapRequest", "ReparentNotify",
		"ConfigureNotify", "ConfigureRequest",
		"GravityNotify", "ResizeRequest",
		"CirculateNotify", "CirculateRequest",
		"PropertyNotify", "SelectionClear",
		"SelectionRequest", "SelectionNotify",
		"ColormapNotify", "ClientMessage",
		"MappingNotify", "LASTEvent"
	};
	
	return names[eventno];
}
