#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <netdb.h>
#include <signal.h>


#include "common.h"
#include "util.h"
#include "port.h"
#include "limits.h"

int     s;			/* the socket */
void	play(int s);
static void usage(void);
char	*extract_host(void);

int
main(int argc, char **argv, char **envp)
{
	struct sockaddr_in sockaddr;
	struct hostent *hp;
	char   *name, *hostname, *cp;
	char    temp[80];
	char   *getenv(const char *);
	void    closem(void);

	s = socket(AF_INET, SOCK_STREAM, 0);
	if (s < 0) {
		perror("socket");
		exit(1);
	}

	if (argc == 1) {
		hostname = getenv("POKER_HOST");
		if (hostname == NULL)
			usage();
	} else
		hostname = argv[1];

	hp = gethostbyname(hostname);
	if (hp == NULL) {
		fprintf(stderr, "gethostbyname(): I can not figure out the "
			" hostname %s\n", hostname);
		if (argc == 1)
			usage();

		fprintf(stderr,"\nI will use the POKER_HOST environmental var");
		hostname = getenv("POKER_HOST");
		if (hostname == NULL)
			usage();
		hp = gethostbyname(hostname);
		if (hp == NULL)
			usage();
	}

	sockaddr.sin_family = AF_INET;
	sockaddr.sin_port = htons(PORT);
	sockaddr.sin_addr.s_addr = *(u_long *) hp->h_addr;

	printf("Waiting for last hand to finish ... ");
	fflush(stdout);
	signal(SIGINT, closem);
	signal(SIGQUIT, closem);

	if (connect(s, &sockaddr, sizeof(sockaddr)) < 0) {
		perror("connect");
		exit(1);
	}
	readln(s, temp);
	if (strcmp(temp, "TOOMANY\n") == 0)
		printf("Table Full!\n");
	else {
		if (strcmp(temp, "OK\n") != 0) {/* must go to another socket */
			close(s);
			sockaddr.sin_port = htons(atoi(temp));
			s = socket(AF_INET, SOCK_STREAM, 0);

			if (connect(s, &sockaddr, sizeof(sockaddr)) < 0) {
				perror("second connect");
				exit(1);
			}
		}
		signal(SIGINT, SIG_IGN);
		signal(SIGQUIT, SIG_IGN);

		if ((name = getenv("POKER")) == NULL)
			if ((name = getenv("NAME")) == NULL)
				name = getenv("USER");

		cp = extract_host();
		sprintf(temp, "%s@%s", name, cp);
		writeln(s, temp);
		sprintf(temp, "%d", getuid());
		writeln(s, temp);
		play(s);
		readln(s, temp);

		while (temp[0] != ':') {
			printf("%s", temp);
			readln(s, temp);
		}
	}
	close(s);
	return 0;
}

static void
usage(void)
{
	static char *msgs[] = {
		"usage : poker poker.host.add.ress",
		"another usage :\tsetenv POKER_HOST poker.host.add.ress",
		"\t\tpoker",
		"info : You can change your name in poker window by changing",
		"       the POKER, or NAME, or USER environmental variable", 
		NULL
	};
	char	**cpp;

	for (cpp = msgs; cpp[0]; cpp++)
		fprintf(stderr, "%s\n", cpp[0]);
	exit(1);
}

/*
 * In retrospect, I think I have successfully proved that I have a high
 * stupidity quantum number.  You can understand my words by examining
 * following function char *extract_host_with_great_studidity()
 * ---hunsoo@sensor.kaist.ac.kr
 */
#if COMMENTED
char *
extract_host_with_great_stupidity(int sock) /* by hunsoo */
{
	struct sockaddr_in saddr;
	int	len;
	struct hostent *hp;
	char	*p;

	len = sizeof(saddr);
	if (getpeername(sock, &saddr, &len))
		err("%s(): ", __F);
	hp = gethostbyaddr(&saddr.sin_addr, len, AF_INET);
	if (hp == NULL) {
		p = inet_ntoa(saddr.sin_addr);
	} else {
		p = hp->h_name;
	}
	return p;
}

#endif

char *
extract_host(void)
{
	static char myname[MAXHOSTNAMELEN];
	char	*p;

	p = myname;
	if (gethostname(myname, MAXHOSTNAMELEN))
		p = "who.am.i";

	return p;
}
