/*
 * Copyright (c) 1997 Hong Hunsoo of (KAIST)
 * The Korea Advanced Institute of Science and Technology
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Neither the name of the KAIST nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <strings.h>
#include <ctype.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xmu/Xmu.h>

#include "common.h"
#include "xcommon.h"
#include "xp_window.h"
#include "cards.h"
#include "limits.h"
#include "util.h"

Window	rootwin;
Window	appwin, dummy_win;
int	screenno;
Display	*dpy;

unsigned long pixel_white, pixel_black;
unsigned long pixel_red, pixel_green, pixel_blue, pixel_yellow;
GC	app_gc, app_gcr, gc_blue, gc_red, gc_yellow, gc_white, gc_black;
GC	gcf;
char	line[256];

int	win_x, win_y;
unsigned int win_height, win_width, win_borderWidth, win_depth;
int	app_x, app_y;
unsigned int app_width, app_height, app_borderWidth, app_depth;
int	dummy;


#define MSG_NO 25
#define PMSG_NO 9

struct msg_t {
	int	m_ycoor;
	char	*m_msg;
};

struct positus {
	int	x, y;
};

struct positus gmsg;	/* the first location of the global msg */

struct msg_t msgs[MSG_NO], perm_msgs[PMSG_NO], *msgt;
extern int s;
int	curr_pot;

static void	xput_str(int y, int x, const char *msg);
static void	xp_clear(void);
void	display_commands(void);
void	make_commands(void);
void	xp_show_choice2nd(char choice);
void	xp_get_string(char *ret);
void	display_pot(void);
void	for_expose(void);
void	get_user_changes(int source, int *reject, int *exchange);
void	display_msgs(void);

int
xp_initial(void)
{
	int	i;
	int	yincre;
	XGCValues xgcv;
	Colormap cmp;
	unsigned long mask;
	Font	font;


	dpy = XOpenDisplay(NULL);
	XSetErrorHandler(error);

	rootwin   = XDefaultRootWindow(dpy);
	screenno  = XDefaultScreen(dpy);
	cmp	  = XDefaultColormap(dpy, screenno);

	pixel_black = XBlackPixel(dpy, screenno);
	pixel_white = XWhitePixel(dpy, screenno);
	pixel_red   = name_to_pixel(cmp, "red");
	pixel_green = name_to_pixel(cmp, "green");
	pixel_blue  = name_to_pixel(cmp, "blue");
	pixel_yellow  = name_to_pixel(cmp, "yellow");

	XGetGeometry(dpy, rootwin, &dummy_win, &win_x, &win_y,
		&win_width, &win_height, &win_borderWidth, &win_depth);

	appwin = XCreateSimpleWindow(dpy, rootwin, 0, 0,
		900, 800, 0, pixel_black, pixel_white);
	XStoreName(dpy, appwin, "network poker");

	XGetGeometry(dpy, appwin, &dummy_win, &app_x, &app_y,
		&app_width, &app_height, &app_borderWidth, &app_depth);
	XSelectInput(dpy, appwin, ExposureMask |KeyPressMask);;

	font = XLoadFont(dpy,
		"-adobe-courier-bold-r-normal--18-180-75-75-m-110-iso8859-1");
	xgcv.font = font;
	xgcv.function = GXcopy;
	xgcv.line_width = 3;
	xgcv.fill_rule = WindingRule;
	xgcv.fill_style = FillSolid;
	xgcv.join_style = JoinRound;

	mask = GCFunction | GCForeground | GCBackground | GCFillRule
		| GCFillStyle | GCLineWidth | GCFont | GCJoinStyle;

	xgcv.foreground = pixel_black;
	xgcv.background = pixel_white;
	app_gc = XCreateGC(dpy, appwin, mask, &xgcv);

	XSetGraphicsExposures(dpy, app_gc, True);

	xgcv.foreground = pixel_white;
	xgcv.background = pixel_black;
	app_gcr = XCreateGC(dpy, appwin, mask, &xgcv);

	xgcv.foreground = pixel_blue;
	xgcv.background = pixel_white;
	gc_blue = XCreateGC(dpy, appwin, mask, &xgcv);

	xgcv.foreground = pixel_red;
	xgcv.background = pixel_white;
	gc_red = XCreateGC(dpy, appwin, mask, &xgcv);

	xgcv.foreground = pixel_yellow;
	xgcv.background = pixel_white;
	gc_yellow = XCreateGC(dpy, appwin, mask, &xgcv);

	xgcv.foreground = pixel_white;
	xgcv.background = pixel_black;
	gc_white = XCreateGC(dpy, appwin, mask, &xgcv);

	xgcv.foreground = pixel_yellow;
	xgcv.background = pixel_white;
	gcf  = XCreateGC(dpy, appwin, mask, &xgcv);

	xgcv.foreground = pixel_black;
	xgcv.background = pixel_white;
	gc_black = XCreateGC(dpy, appwin, mask, &xgcv);

	XMapWindow(dpy, appwin);
	card_init(dpy, screenno);

	XFlush(dpy); /* XXX */

	yincre = 60;
	for (i=0; i < PMSG_NO; i++) {
		perm_msgs[i].m_msg = (char *) malloc(100);
		if (perm_msgs[i].m_msg == NULL)
			err("%s(): msgs[%d].m_msg == NULL", __F, i);
		perm_msgs[i].m_ycoor = yincre;
		perm_msgs[i].m_msg[0] = 0;
		yincre += 20;
	} 
	make_commands();

	yincre -= 20;
	for (i=0; i < MSG_NO; i++) {
		msgs[i].m_msg = (char *) malloc(100);
		if (msgs[i].m_msg == NULL)
			err("%s(): msgs[%d].m_msg == NULL", __F, i);
		msgs[i].m_msg[0] = 0;
		msgs[i].m_ycoor = yincre;
		yincre += 20;
	}

	msgt = msgs;
	return 0;
}

void
xp_clear_mid(void)
{
	int	i, offset;

	offset = gmsg.x;

	XClearArea(dpy, appwin,
		offset, 0, app_width - offset, app_height, 0);
	for (i=0; i<MSG_NO; i++)
		msgs[i].m_msg[0] = 0;
	msgt = msgs;
}



void
really(void)
{
	return;
}
	

void
xp_get_input(int des, char *val)
{
	XEvent event;
	KeySym mykey;
	int	no;
	char	buf[4];

	do {
		XNextEvent(dpy, &event);

		switch (event.type) {
		case Expose:
			for_expose();
			break;
		case KeyPress:
			no = XLookupString((XKeyEvent *) &event,
				buf, 4, &mykey, 0);
			if (no == 1 && buf[0] == 'q')
				really();
			break;
		}
		*val = buf[0];
	} while (*val == '\022');
}

void
xp_show_choice2nd(char choice)
{
	char	buf[10];
	sprintf(buf, "%c", choice);
	xp_appendmsg(buf);
	XFlush(dpy);
}

void
xp_show_choice(char choice, int count)
{
	char	buf[10];
	sprintf(buf, "%c ", choice);
	xp_appendmsg(buf);
	XFlush(dpy);
}


void
xp_putmsg(const char *fmt, ...)
{
	char	tmp0[100];
	va_list cp;
	va_start(cp, fmt);

	if (msgt == &msgs[MSG_NO]) {
		strcpy(tmp0, (--msgt)->m_msg);
		xp_clear_mid();
		strcpy(msgt->m_msg, tmp0);
		msgt++;
		display_commands();
		display_msgs();
		display_pot();
	}

	vsprintf(msgt->m_msg, fmt, cp);
	XDrawString(dpy, appwin, app_gc, gmsg.x, msgt->m_ycoor,
		msgt->m_msg, strlen(msgt->m_msg));

	msgt++;
	XFlush(dpy);
	va_end(cp);
}

void
xp_appendmsg(const char *string)
{
	strcat(msgt->m_msg, string);
	XDrawString(dpy, appwin, app_gc, gmsg.x, msgt->m_ycoor,
		msgt->m_msg, strlen(msgt->m_msg));

	XFlush(dpy);
}
void
xp_prevdeletemsg(void)
{
	if (msgt == &msgs[0])
		return;
	msgt--;
	XClearArea(dpy, appwin, gmsg.x, msgt->m_ycoor - 14,
		app_width - gmsg.x, msgt->m_ycoor + 4, 0);
	msgt->m_msg[0] = 0;

	XFlush(dpy);
}


void
xp_deletemsg(void)
{
	XClearArea(dpy, appwin, gmsg.x, msgt->m_ycoor - 14,
		app_width - gmsg.x, msgt->m_ycoor + 4, 0);
	msgt->m_msg[0] = 0;

	XFlush(dpy);
}

void
display_msgs(void)
{
	int	i;
	struct msg_t *msg;

	for (i=0; i<MSG_NO; i++) {
		msg = &msgs[i];
		if (msg->m_msg[0])
			XDrawString(dpy, appwin, app_gc, gmsg.x, msg->m_ycoor,
				msg->m_msg, strlen(msg->m_msg));
	}
	return;
}

struct card_t {
	int	c_suit, c_rank;
	int	c_x, c_y;
};

struct persona_t {
	int	ps_mytern;
	char	ps_name[128];
	char	ps_msg[128];
	int	ps_property, ps_betting;
	struct card_t ps_cards[5];
	struct positus ps_mpositus;
};

struct persona_t personas[5], *psp;
static int plus, playerno;

#define UNIT 45

void
persona_init(void)
{
	int	i, j;
	int	offx, offy;
	struct card_t *cp;

	offx = 10 + 5 * UNIT;
	for (i=0; i < 5; i++) { /* cards locations */
		offy = 10 + i * 130;
		for (j=0; j < 5; j++) { /* for cards of this persona_t */
			cp = &(personas[i].ps_cards[j]);
			cp->c_x = 10 + UNIT * j;
			cp->c_y = offy;
			cp->c_suit = 4; /* backface */
			cp->c_rank = 0;
		}

	}

	if (UNIT < CARD_WIDTH)
		plus = CARD_WIDTH - UNIT;
	else
		plus = 0;

	offx = 10 + 4 * UNIT + CARD_WIDTH + 5;
	gmsg.x = offx + 200;
	gmsg.y = 20;

	for (i=0; i < 5; i++) { /* message locations */
		offy = 10 + i * 130;
		psp = &personas[i];

		psp->ps_mytern = 0;
		psp->ps_mpositus.x = offx;
		psp->ps_mpositus.y = offy;
		psp->ps_property = 0;
		psp->ps_betting = 0;
		psp->ps_name[0] = 0;
		psp->ps_msg[0] = 0;
	}
	curr_pot = 0;

	return;
}

/*
 * oops! cards.c and pokerd.c has different suit system
 */
static int trans[4] = {2, 3, 1, 0};

void
handle_cardstr(struct persona_t *tps, char *cardsstr)
{
	char	*p;
	int	i, suit, rank;
	struct card_t *cp;

	cp = &tps->ps_cards[0];
	for (p = cardsstr, i = 0; i < 5; i++, cp++, p += 3) {  
		rank = p[0] - '0';
		suit = p[1] - '0';
		if (rank < 0 || rank >= 13 ||suit < 0 || suit > 3)
			err("%s(): FATAL ERROR!!", __F);

		cp->c_suit = trans[suit];
		cp->c_rank = rank + 1;
		if (cp->c_rank == 13)
			cp->c_rank = 0;
	}
	return;
}

void
display_usercards(struct persona_t *tps)
{
	struct card_t *cp;
	int	i;

	if (tps->ps_name[0] == 0)
		return;

	for (i=0, cp = &tps->ps_cards[0]; i < 5; i++, cp++)
		put_card(appwin, cp->c_suit, cp->c_rank, cp->c_x, cp->c_y);
	XFlush(dpy);
	return;
}

void
display_usermsg(struct persona_t *tps)
{
	char	tmp0[64], tmp1[64], *p;
	int	xoff, yoff;

	if (tps->ps_name[0] == 0)
		return;
	xoff =  tps->ps_mpositus.x;
	yoff =  tps->ps_mpositus.y;

	XClearArea(dpy, appwin, xoff, yoff, 200, 130, 0);

	if (tps->ps_mytern)
		XDrawArc(dpy, appwin, app_gc,
			xoff+2, yoff + 2, 13, 13, 0, 64 * 360);

	sprintf(tmp0, "%s", tps->ps_name);
	p = index(tmp0, '@');

	yoff += 20;
	if (p) {
		strcpy(tmp1, p); /* contains '@' */
		p[0] = 0;
		strcat(tmp0, " at");
		XDrawString(dpy, appwin, app_gc,
			xoff + 20, yoff, tmp0, strlen(tmp0));
		yoff += 20;
		tmp1[17] = 0; /* to shorten a long name */
		XDrawString(dpy, appwin, app_gc, xoff, yoff, tmp1, strlen(tmp1));
	} else {
		XDrawString(dpy, appwin, app_gc, xoff+20, yoff, tmp0, strlen(tmp0));
	}

	yoff += 20;
	sprintf(tmp0, "%-5d", tps->ps_property);
	XDrawString(dpy, appwin, app_gc, xoff, yoff, tmp0, strlen(tmp0));

	sprintf(tmp0, "bet:%-3d", tps->ps_betting);
	XDrawString(dpy, appwin, app_gc, xoff + 100, yoff, tmp0, strlen(tmp0));

	p = index(tps->ps_msg, '@');
	yoff += 20;
	if (p) {
		*p++ = 0;
		XDrawString(dpy, appwin, app_gc,
			xoff, yoff, tps->ps_msg, strlen(tps->ps_msg));
		yoff += 20;
		XDrawString(dpy, appwin, app_gc, xoff, yoff, p, strlen(p));
	} else {
		XDrawString(dpy, appwin, app_gc,
			xoff, yoff, tps->ps_msg, strlen(tps->ps_msg));
	}

	XFlush(dpy);
	return;
}

void
for_ternchange(void)
{
	struct persona_t *tps;
	int	i;

	for (i=0; i<5; i++) {
		tps = &personas[i];
		display_usermsg(tps);
	}
}

void
for_expose(void)
{
	struct persona_t *tps;
	int	i;

	for (i=0; i<5; i++) {
		tps = &personas[i];
		display_usercards(tps);
		display_usermsg(tps);
	}
	display_commands();
	display_msgs();
	display_pot();
	XFlush(dpy);
	return;
}

void
another(void)
{
	int	i, j;
	struct card_t *cp;

	curr_pot = 0;
	for (i=0; i < 5; i++) { /* cards locations */
		psp = &personas[i];
		for (j=0; j < 5; j++) { /* for cards of this persona_t */
			cp = &(psp->ps_cards[j]);
			cp->c_suit = 4; /* backface */
			cp->c_rank = 0;
		}
		psp->ps_msg[0] = 0;
		psp->ps_betting = 0;

		display_usercards(psp);
		display_usermsg(psp);
		display_pot();
	}
	xp_clear_mid();
	display_commands();
	XFlush(dpy);
	return;
}

void
make_commands(void)
{
	msgt = &perm_msgs[0];
	strcpy(msgt->m_msg, "COMMANDS");
	msgt++; strcpy(msgt->m_msg, "C) Call last raise");
	msgt++; strcpy(msgt->m_msg, "F) Fold this hand");
	msgt++; strcpy(msgt->m_msg, "P) Pass (Check)");
	msgt++; strcpy(msgt->m_msg, "Q) Quit after this hand");
	msgt++; strcpy(msgt->m_msg, "R) Raise last bet");
	msgt++; strcpy(msgt->m_msg, "S) Sit out for a while");
	msgt++; strcpy(msgt->m_msg, "---------");
	msgt++; msgt->m_msg = NULL;
}

void
display_commands(void)
{
	struct msg_t *mt;

	for (mt = &perm_msgs[0]; mt->m_msg; mt++)
		XDrawString(dpy, appwin, app_gc, gmsg.x, mt->m_ycoor,
			mt->m_msg, strlen(mt->m_msg));
	XFlush(dpy);
	return;
}

void
display_pot(void)
{
	char	tmp[32];

	sprintf(tmp, "----> POT: %-6d", curr_pot);
	XClearArea(dpy, appwin, gmsg.x, 0, app_width - gmsg.x, 25, 0);
	XDrawString(dpy, appwin, app_gc, gmsg.x, 20, tmp, strlen(tmp));
	return;
}


/*
 * Listen to poker daemon and update the screen as it says. These are the
 * commands from the daemon to this program: *
 * 
 *	Nn	name of player n
 *	Cn	cards for player n
 *	P       pot value
 * 	U	update screen
 *	X	clear screen
 *	Hn	action for player n (fold, raise, call, etc.)
 *	Rn	result for player n (straight, two pair, --fold--, etc.)
 *	Bn	total amount bet by player n this hand
 *	$n	amount of cash player n has
 *	D	get discards from this player
 *	?	get bet from this player - amount to call follows
 * 	Q	player has quit; cleanup & terminate
 *
 *	W	player name following has won this hand
 *	T	tells which player's turn it is; cursor is put next to his name
 *	M	message to print on bottom of screen 
 * 
 *
 * These are the commands from this program to the daemon:
 * 
 *	Q	quit after this hand is finished
 *	F	fold
 *
 * Dxxxxx	discard these cards (k=keep,d=discard)
 *	C	call
 *
 * Rxx		raise by this amount
 *	P	pass - not allowed if player has to call someone else's raise
 *	S	toggles sitting out switch
 *	A	Announce a new player (kl@unido) *
 * 
 */

void
play(int s)
{
	char    temp[80];	/* buffer to read command into */
	char    temp2[80];	/* outgoing commands are sometimes built here */
	int     n;		/* determine which player we're talking about */
	int     minimum;	/* minimum bet to stay in */
	char    c;		/* player's command */
	int     done;		/* get commands until something valid */
	int     quit = FALSE;	/* has player quit yet? */
	int     out = FALSE;	/* is player sitting out? */
	char    cards[20];	/* hold player's current hand */
	int	rejected_no;
	int	rejected[10];
	int     i, x, raise_cnt;
	struct persona_t *ternp;

	xp_initial();
	xp_first();
	persona_init();
	display_commands();
	playerno = 2;


	ternp = &personas[0];
	while (!quit) {
		readln(s, temp);
		n = temp[1] - '0';	/* which player? */
		if (n > playerno)
			playerno = n;
		psp = &personas[n];

		switch (temp[0]) {
		case 'M':
			xput_str(15, 40, temp + 1);
			XFlush(dpy);
			break;
		case 'N':
			strcpy(psp->ps_name, temp + 2);
			display_usermsg(psp);
			display_usercards(psp);
			break;
		case 'T':
			ternp->ps_mytern = 0;
			psp->ps_mytern = 1;
			for_ternchange();
			ternp = psp;
			break;
		case 'C':
			handle_cardstr(psp, temp + 2);
			display_usercards(psp);
			strcpy(cards, temp + 2);
			break;
		case 'U':
			break;
		case 'X':
			xp_clear();
			break;
		case 'P':
			curr_pot = atoi(temp + 1);
			display_pot();
			break;
		case 'B':
			psp->ps_betting = atoi(temp + 2);
			display_usermsg(psp);
			break;
		case 'H':
			strcpy(psp->ps_msg, temp+2);
			display_usermsg(psp);
			break;
		case 'R':
			strcpy(psp->ps_msg, temp+2);
			display_usermsg(psp);
			break;
		case '$':
			psp->ps_property =  atoi(temp + 2);
			display_usermsg(psp);
			break;
		case 'Q':
			/* endwin(); */
			quit = TRUE;
			break;

		case 'W':
			temp[strlen(temp) - 1] = 0;	/* strip newline */
			sprintf(temp2, ">>>>%s<<<<", temp + 1);
			xp_putmsg(temp2);
			xp_putmsg("wins this hand!");
			xp_putmsg("--Hit space bar!!");
			XFlush(dpy);

			xp_get_input(1234, &c);
			while (c != ' ')
				xp_get_input(1234, &c);

			if (out) {
				xp_putmsg("Hit space bar to come back!!");
				xp_get_input(1234, &c);
				while (c != ' ')
					xp_get_input(1234, &c);
				xp_putmsg("Waiting...");
				writeln(s, "S");
				out = FALSE;
			}
			another();

			break;
		case 'D':
			/* find out what player wants to discard */
			get_user_changes(1234, &rejected_no, rejected);

			i = 0;
			strcpy(temp2, "Dkkkkk");
			for (i=0; i<rejected_no; i++) {
				x = rejected[i]+1;
				if (x > 0 && x < 6)
					temp2[x] = 'd';
			}
			writeln(s, temp2);
			break;
		case '?': /* user input---hunsoo */
			/* time to let player bet on this hand */
			minimum = atoi(temp + 1);
			done = FALSE;

			xp_putmsg("Call/Fold/Pass/Quit/Raise/Sitout??");
			while (done == FALSE) {
				XFlush(dpy);
				xp_get_input(1234, &c);
				if (islower(c))
					c = toupper(c);

				done = TRUE;
				switch (c) {
				case 'Q':
					done = FALSE;
				case 'C':
					if (c == 'C' && minimum == 0) {
						xp_putmsg("No raise to call; "
							"pass or raise.");
						XFlush(dpy);
						done = FALSE;
						break;
					}
				case 'F':
					sprintf(temp, "%c", c);
					writeln(s, temp);
					if (c == 'C')
						strcpy(temp, "Call!");
					else if (c == 'Q') {
						strcpy(temp, "Quitting.");
						strcat(temp,
						"(ie This is last round)");
					} else
						strcpy(temp, "Fold!");
					xp_putmsg(temp);
					break;
				case 'S':
					done = FALSE;
					out = TRUE;
					writeln(s, "S");
					xp_putmsg("Sitting out.");
					break;
				case 'P':
					if (minimum > 0) {
						xp_putmsg("Can't pass; "
						"meet last raise or fold.");
						XFlush(dpy);
						done = FALSE;
					} else {
						writeln(s, "P");
						xp_putmsg("Pass!");
					}
					break;
				case 'R':
					strcpy(temp, "0");
					xp_putmsg("Raise!! How much money"
						"(0-50)?");
					raise_cnt = 0;
					while (atoi(temp) <= 0) {
						if (raise_cnt) {
							xp_prevdeletemsg();
							xp_putmsg("Wrong input!"
								" try again..");
						}
						raise_cnt = 1;
						xp_get_string(temp);
						if (atoi(temp) > RAISE_LIMIT) {
							xp_putmsg(
							"Raise limit = %d and "
							"your input = %d",
							RAISE_LIMIT,atoi(temp));
							xp_putmsg("Please "
							"raise again..");
							strcpy(temp, "0");
						}
					}
					minimum = atoi(temp);
					sprintf(temp, "R%d", minimum);
					writeln(s, temp);
					break;
				default:
					done = FALSE;
					/* try again */
					break;
				}
			}
			break;
		default:
			/* ignore unknown commands */
			break;
		}
	}
	printf("\n\n\n");
}


void 
closem(void)
{
	close(s);
	exit(1);
}

static void
xput_str(int y, int x, const char *msg)
{
	XDrawString(dpy, appwin, app_gc, 20 * x, 20 * y, msg, strlen(msg));
	return;
}

static void
xp_clear(void)
{
	return;
}

void
get_user_changes(int source, int *reject, int *exchange)
{
	int     no_of_cards, complete, count, ok, value;
	char    choice;

	xp_putmsg("Which cards do you");
	xp_putmsg("want to reject(1-5)?");

	no_of_cards = 0;
	complete = 0;
	choice = ' ';

	do {
		do
			xp_get_input(source, &choice);
		while ((choice < '1' || choice > '5') && choice != 13 &&
		       choice != '\n' && choice != '\b' && choice != '\177') ;

		switch (choice) {
		case '\n':
		case 13:
			complete = 1;
			*reject = no_of_cards;
			break;
		case '\b':
		case '\177':
			xp_deletemsg();
			no_of_cards = 0;
			break;
		default:
			if (no_of_cards < 6) {
				value = choice - '1';
				ok = 1;
				for (count = 0; count < no_of_cards; count++)
					if (value == exchange[count])
						ok = 0;

				if (ok) {
					xp_show_choice(choice, no_of_cards);
					exchange[no_of_cards++] = value;
				}
			}
		}		/* switch */
	} while (!complete);
	xp_putmsg("");
}

void
xp_get_string(char *ret)
{
	int     complete;
	char    choice, *p;

	complete = 0;
	choice = ' ';

	p = ret;
	do {
		xp_get_input(1243, &choice);

		switch (choice) {
		case '\n':
		case 13:
			complete = 1;
			break;
		case '\b':
		case '\177':
			xp_deletemsg();
			p = ret;
			break;
		default:
			xp_show_choice2nd(choice);
			*p++ = choice;
		}		/* switch */
	} while (!complete);
	*p = 0;
	xp_putmsg("");
}
